""" Routes for managing roles. """

# pylint: disable=C0116

from flask import render_template, request, redirect, url_for, flash
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import Role
from FabOMatic.database.repositories import RoleRepository
from .webapplication import DBSession, app


@app.route("/roles")
@login_required
def roles():
    session = DBSession()
    roles_list = session.query(Role).all()
    return render_template("view_roles.html", roles=roles_list)


@app.route("/roles/add", methods=["GET", "POST"])
@login_required
def add_role():
    if request.method == "POST":
        session = DBSession()
        role_name = request.form["role_name"]
        authorize_all = request.form.get("authorize_all", "off") == "on"
        maintenance = request.form.get("maintenance", "off") == "on"
        backend_admin = request.form.get("backend_admin", "off") == "on"
        role = Role(
            role_name=role_name,
            authorize_all=authorize_all,
            maintenance=maintenance,
            reserved=False,
            backend_admin=backend_admin,
        )
        session.add(role)
        session.commit()
        return redirect(url_for("roles"))
    else:
        return render_template("add_role.html")


@app.route("/roles/edit/<int:role_id>", methods=["GET", "POST"])
@login_required
def edit_role(role_id):
    session = DBSession()
    role = session.query(Role).filter_by(role_id=role_id).one()

    # Block editing of reserved roles
    if not role:
        flash(gettext("Role not found."), "error")
        return redirect(url_for("roles"))

    if role.reserved:
        flash(gettext("Cannot edit reserved role."), "error")
        return redirect(url_for("roles"))

    if request.method == "POST":
        role.role_name = request.form["role_name"]
        role.authorize_all = request.form.get("authorize_all", "off") == "on"
        role.maintenance = request.form.get("maintenance", "off") == "on"
        role.backend_admin = request.form.get("backend_admin", "off") == "on"
        session.add(role)
        session.commit()
        return redirect(url_for("roles"))

    return render_template("edit_role.html", role_id=role.role_id, role=role)


@app.route("/roles/delete/<int:role_id>", methods=["GET", "POST"])
def delete_role(role_id):
    session = DBSession()
    role_repo = RoleRepository(session)
    role = role_repo.get_by_id(role_id)
    if not role:
        flash(gettext("Role not found."), "error")
        return redirect(url_for("roles"))

    if role.reserved:
        flash(gettext("Cannot edit reserved role."), "error")
        return redirect(url_for("roles"))

    if request.method == "POST":
        role_repo.delete(role)
        flash(gettext("Role deleted successfully."))
        return redirect(url_for("roles"))

    return render_template("delete_role.html", role=role)



"""Main module of the backend."""

import logging
import threading
from time import sleep

from FabOMatic.database.DatabaseBackend import DatabaseBackend
from FabOMatic.mqtt.MQTTInterface import MQTTInterface
from FabOMatic.logic.MsgMapper import MsgMapper
from FabOMatic.logger import configure_logger


class Backend:
    """Backend class."""

    def __init__(self, config_file: str = None):
        if config_file is None:
            self._db = DatabaseBackend()
            self._mqtt = MQTTInterface()
        else:
            self._db = DatabaseBackend(config_file)
            self._mqtt = MQTTInterface(config_file)

        self._mapper = MsgMapper(self._mqtt, self._db)
        self._mapper.registerHandlers()
        self._flaskThread = None

    def connect(self) -> bool:
        """Connect to the MQTT broker and the database."""
        try:
            session = self._db.getSession()
            logging.info(f"Session info: {session.info}")
            self._db.createAndUpdateDatabase()
            self._mqtt.connect()
            return True
        except Exception as ex:
            logging.error("Connection failed: %s", ex, exc_info=True)
            return False

    @property
    def connected(self) -> bool:
        """Check if the MQTT broker is connected."""
        return self._mqtt.connected

    def disconnect(self):
        """Disconnect from the MQTT broker"""
        self._mqtt.disconnect()

    def publishStats(self):
        self._mqtt.publishStats()


_flaskThread: threading.Thread = None


def _startApp() -> None:
    from FabOMatic.web.webapplication import app

    app.run(host="0.0.0.0", port=23336, debug=True, use_reloader=False, ssl_context="adhoc")


def startServer() -> None:
    global _flaskThread
    _flaskThread = threading.Thread(target=lambda: _startApp(), daemon=True)
    _flaskThread.start()


def start(loglevel):
    """Main function of the backend."""
    configure_logger(loglevel)
    logging.info("Starting backend...")
    back = Backend()
    startServer()

    while True:
        if not back.connected:
            if not back.connect():
                logging.error("Failed to connect to Database or MQTT broker")
        else:
            back.publishStats()
        sleep(5)


if __name__ == "__main__":
    start(10)



""" This module contains the routes for the authorizations. """

# pylint: disable=C0116

from flask import flash, render_template, request, redirect, url_for
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import Authorization, Machine, User
from .webapplication import DBSession, app, excel


@app.route("/authorizations", methods=["GET"])
@login_required
def view_authorizations():
    user_filter = request.args.get("user")
    machine_filter = request.args.get("machine")

    with DBSession() as session:
        query = (
            session.query(Authorization).join(Authorization.user).filter_by(deleted=False).join(Authorization.machine)
        )

        if user_filter and user_filter.isdigit():
            query = query.filter(User.user_id == int(user_filter))
        if machine_filter and machine_filter.isdigit():
            query = query.filter(Machine.machine_id == int(machine_filter))

        authorizations = query.all()
        users = session.query(User).filter_by(deleted=False).all()
        machines = session.query(Machine).all()
        return render_template(
            "view_authorizations.html", authorizations=authorizations, users=users, machines=machines
        )


@app.route("/authorizations/add", methods=["GET"])
@login_required
def add_authorization():
    with DBSession() as session:
        users = session.query(User).filter_by(deleted=False).all()
        machines = session.query(Machine).all()
        return render_template("add_authorization.html", users=users, machines=machines)


@app.route("/authorizations/create", methods=["POST"])
@login_required
def create_authorization():
    with DBSession() as session:
        authorization_data = request.form
        new_authorization = Authorization(
            user_id=authorization_data["user_id"],
            machine_id=authorization_data["machine_id"],
        )
        session.add(new_authorization)
        try:
            session.commit()
        except Exception as e:
            flash(str(e), "danger")
        return redirect(url_for("view_authorizations"))


@app.route("/authorizations/edit/<int:authorization_id>", methods=["GET"])
@login_required
def edit_authorization(authorization_id):
    with DBSession() as session:
        authorization = session.query(Authorization).filter_by(authorization_id=authorization_id).one()
        users = session.query(User).filter_by(deleted=False).all()
        machines = session.query(Machine).all()
        if authorization:
            return render_template(
                "edit_authorization.html", authorization=authorization, users=users, machines=machines
            )
        else:
            return gettext("Authorization not found"), 404


@app.route("/authorizations/update", methods=["POST"])
@login_required
def update_authorization():
    with DBSession() as session:
        authorization_data = request.form
        authorization = (
            session.query(Authorization).filter_by(authorization_id=authorization_data["authorization_id"]).one()
        )
        if authorization:
            authorization.user_id = authorization_data["user_id"]
            authorization.machine_id = authorization_data["machine_id"]
            try:
                session.commit()
            except Exception as e:
                flash(str(e), "danger")
            return redirect(url_for("view_authorizations"))
        else:
            return gettext("Authorization not found"), 404


@app.route("/authorizations/delete/<int:authorization_id>", methods=["GET", "POST"])
@login_required
def delete_authorization(authorization_id):
    with DBSession() as session:
        authorization = session.query(Authorization).filter_by(authorization_id=authorization_id).one()
        if not authorization:
            return "Authorization not found", 404

        if request.method == "POST":
            session.delete(authorization)
            session.commit()
            return redirect(url_for("view_authorizations"))

        return render_template("delete_authorization.html", authorization=authorization)


@app.route("/authorizations/bulkadd", methods=["GET", "POST"])
@login_required
def bulkadd_authorizations():
    with DBSession() as session:
        machines = session.query(Machine).all()
        users = session.query(User).filter_by(deleted=False).order_by(User.user_id).all()

        if request.method == "POST":
            authorization_data = request.form
            machine = session.query(Machine).filter_by(machine_id=authorization_data["machine_id"]).one()
            for user_id in authorization_data.getlist("user_ids"):
                existing = (
                    session.query(Authorization).filter_by(user_id=user_id, machine_id=machine.machine_id).first()
                )
                if existing:
                    continue
                authorization = Authorization(user_id=user_id, machine_id=machine.machine_id)
                session.add(authorization)
            session.commit()
            flash(gettext("Authorizations added successfully"), "success")
            return redirect(url_for("view_authorizations"))

        return render_template("bulkadd_authorizations.html", machines=machines, users=users)


@app.route("/authorizations/export", methods=["GET"])
@login_required
def authorizations_export():
    session = DBSession()
    return excel.make_response_from_tables(session, [Authorization, User, Machine], "xlsx", file_name="authorizations")



import json

from FabOMatic.database.constants import USER_LEVEL


class Parser:
    @staticmethod
    def parse(json_data: str):
        """
        Parses the given JSON data and returns the corresponding query object based on the 'action' field.

        Args:
            json_data (str): The JSON data to parse.

        Returns:
            object: The deserialized query object based on the 'action' field.

        Raises:
            ValueError: If the 'action' field is missing or invalid.
        """
        data = json.loads(json_data)
        if "action" in data:
            match data["action"]:
                case "checkuser":
                    return UserQuery.deserialize(json_data)
                case "checkmachine":
                    return MachineQuery.deserialize(json_data)
                case "startuse":
                    return StartUseQuery.deserialize(json_data)
                case "inuse":
                    return InUseQuery.deserialize(json_data)
                case "stopuse":
                    return EndUseQuery.deserialize(json_data)
                case "maintenance":
                    return RegisterMaintenanceQuery.deserialize(json_data)
                case "alive":
                    return AliveQuery.deserialize(json_data)
                case _:
                    raise ValueError("Invalid action")
        else:
            raise ValueError("Missing action field")


class BaseJson:
    def toJSON(self):
        return json.dumps(self, default=lambda o: (o.__dict__), sort_keys=True, separators=(",", ":"))

    def serialize(self) -> str:
        return json.dumps(self.__dict__)


class UserQuery(BaseJson):
    def __init__(self, card_uid: str):
        self.uid = card_uid
        self.action = "checkuser"

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        return UserQuery(data["uid"])


class MachineQuery(BaseJson):
    def __init__(self):
        self.action = "checkmachine"

    @staticmethod
    def deserialize(json_data: str):
        return MachineQuery()


class AliveQuery(BaseJson):
    def __init__(self, version: str, ip: str, serial: str, heap: int):
        self.action = "alive"
        self.version = version
        self.ip = ip
        self.serial = serial
        self.heap = heap

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        # Serial and Heaps have been added in FW revision 0.6.7
        if data.get("serial") is None:
            data["serial"] = ""
        if data.get("heap") is None:
            data["heap"] = 0
        return AliveQuery(data["version"], data["ip"], data["serial"], data["heap"])


class StartUseQuery(BaseJson):
    def __init__(self, card_uid: str, replay: bool = False):
        self.uid = card_uid
        self.action = "startuse"
        self.replay = replay

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        if data.get("replay") is None:
            data["replay"] = False
        return StartUseQuery(data["uid"], data["replay"])


class EndUseQuery(BaseJson):
    def __init__(self, card_uid: str, duration_s: int, replay: bool = False):
        self.uid = card_uid
        self.duration = duration_s
        self.action = "stopuse"
        self.replay = replay

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        if data.get("replay") is None:
            data["replay"] = False
        return EndUseQuery(data["uid"], data["duration"], data["replay"])


class InUseQuery(BaseJson):
    def __init__(self, card_uid: str, duration_s: int):
        self.uid = card_uid
        self.duration = duration_s
        self.action = "inuse"

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        return InUseQuery(data["uid"], data["duration"])


class RegisterMaintenanceQuery(BaseJson):
    def __init__(self, card_uid: str, replay: bool = False):
        self.uid = card_uid
        self.action = "maintenance"
        self.replay = replay

    @staticmethod
    def deserialize(json_data: str):
        data = json.loads(json_data)
        if data.get("replay") is None:
            data["replay"] = False
        return RegisterMaintenanceQuery(data["uid"], data["replay"])


class UserResponse:
    def __init__(
        self, request_ok: bool, is_valid: bool, holder_name: str, user_level: USER_LEVEL | int, missing_auth: bool
    ):
        self.request_ok = request_ok
        self.is_valid = is_valid
        self.name = holder_name
        self.missing_auth = missing_auth
        if isinstance(user_level, USER_LEVEL):
            self.level = user_level.value
        else:
            self.level = user_level

    def serialize(self) -> str:
        return json.dumps(self.__dict__)


class MachineResponse:
    def __init__(
        self,
        request_ok: bool,
        is_valid: bool,
        needs_maintenance: bool,
        allowed: bool,
        name: str,
        type_id: int,
        timeout_min: int = 0,
        grace_period_min: int = 0,
        description: str = "",
    ):
        self.request_ok = request_ok
        self.is_valid = is_valid
        self.maintenance = needs_maintenance
        self.allowed = allowed
        self.name = name
        self.logoff = timeout_min
        self.type = type_id
        self.grace = grace_period_min
        self.description = description

    def serialize(self) -> str:
        return json.dumps(self.__dict__)


class SimpleResponse:
    def __init__(self, request_ok: bool, message: str = ""):
        self.request_ok = request_ok
        self.message = message

    def serialize(self) -> str:
        return json.dumps(self.__dict__)

    def __str__(self):
        return self.serialize()



"""This module contains constants used in the database module"""

from enum import Enum


class USER_LEVEL(Enum):
    """This class maps to the Board c++ FabUser.UserLevel int value"""

    INVALID = 0
    NORMAL = 1
    ADMIN = 2


DEFAULT_TIMEOUT_MINUTES = 24 * 60
DEFAULT_GRACE_PERIOD_MINUTES = 2



""" A module for the MQTTInterface class. """

import logging
import os
import json
from time import sleep
import toml
import paho.mqtt.client as mqtt
from .mqtt_types import Parser

MODULE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
CONFIG_FILE = os.path.join(MODULE_DIR, "conf", "settings.toml")


class MQTTInterface:
    """
    A class representing an MQTT interface.

    Attributes:
        _settings_path (str): The path to the MQTT settings file.
        _broker (str): The MQTT broker address.
        _port (int): The MQTT broker port.
        _client_id (str): The client ID for connecting to the MQTT broker.
        _topic (str): The MQTT topic for publishing queries.
        _reply_subtopic (str): The MQTT subtopic for publishing replies.
        _statsTopic (str): The MQTT topic for publishing statistics.
        _messageCallback (callable): The callback function for processing received messages.
        _connected (bool): Indicates whether the MQTT interface is connected to the broker.
        _handlers (dict): A dictionary of message handlers.
        _msg_send_count (int): The count of sent messages.
        _msg_recv_count (int): The count of received messages.
    """

    def __init__(self, path=CONFIG_FILE):
        """
        Initializes an instance of the MQTTInterface class.

        Args:
            path (str, optional): The path to the MQTT settings file. Defaults to CONFIG_FILE.
        """
        self._settings_path = path
        self._messageCallback = None
        self._connected = False
        self._handlers = {}
        self._loadSettings()
        self._msg_send_count = 0
        self._msg_recv_count = 0

    def _loadSettings(self) -> None:
        """
        Loads the MQTT settings from the settings file.
        """
        settings = toml.load(self._settings_path)["MQTT"]
        self._broker = settings["broker"]
        self._port = settings["port"]
        self._client_id = settings["client_id"]
        self._topic = settings["topic"]
        self._reply_subtopic = settings["reply_subtopic"]
        self._statsTopic = settings["stats_topic"] + "/" + self._client_id
        logging.info("Loaded MQTT settings from file %s", self._settings_path)

    def _extractMachineFromTopic(self, topic: str) -> str:
        """
        Extracts the machine name from the MQTT topic.

        Args:
            topic (str): The MQTT topic.

        Returns:
            str: The machine name extracted from the topic.
        """
        elems = topic.split("/")
        if len(elems) < 2:
            return None
        if elems[0].capitalize() != self._topic.capitalize():
            return None

        return elems[1]

    def _onMessage(self, *args):
        """
        Callback function for handling received MQTT messages.

        Args:
            *args: Variable length argument list.
        """
        topic = args[2].topic
        message = args[2].payload.decode("utf-8")
        self._msg_recv_count += 1

        machine = self._extractMachineFromTopic(topic)
        if not machine:
            logging.warning("Could not extract machine from topic : %s", topic)
            return

        if self._messageCallback is None:
            logging.warning("No message callback set, message will be ignored")
            return

        try:
            query = Parser.parse(message)
            if query is not None and machine.isdigit():
                self._messageCallback(int(machine), query)
        except ValueError:
            logging.warning("Invalid message received: %s on machine %s", message, machine)
            return

    def publishQuery(self, machine: str, message: str) -> bool:
        """
        Publishes a query message to a specific machine.

        Args:
            machine (str): The name of the machine.
            message (str): The query message.

        Returns:
            bool: True if the message was published successfully, False otherwise.
        """
        return self._publish(f"{self._topic}{machine}", message)

    def publishReply(self, machine: str, message: str) -> bool:
        """
        Publishes a reply message to a specific machine.

        Args:
            machine (str): The name of the machine.
            message (str): The reply message.

        Returns:
            bool: True if the message was published successfully, False otherwise.
        """
        self._msg_send_count += 1
        return self._publish(f"{self._topic}/{machine}/reply", message)

    def _publish(self, topic: str, message: str) -> bool:
        """
        Publishes a message to the MQTT broker.

        Args:
            topic (str): The MQTT topic.
            message (str): The message to publish.

        Returns:
            bool: True if the message was published successfully, False otherwise.
        """
        if self.connected:
            result = self._client.publish(topic, message)
            logging.debug("Publishing %s : %s, result: %s", topic, message, result)
            return True
        logging.error("Not connected to MQTT broker %s", self._broker)
        return False

    def _onDisconnect(self, client, userdata, rc):
        """
        Callback function for handling MQTT disconnection.

        Args:
            client: The MQTT client.
            userdata: The user data.
            rc (int): The reason code for disconnection.
        """
        if self._connected:
            logging.info("Disconnected to MQTT broker reason code:%d", rc)
        self._connected = False
        self._client.loop_stop()

    def _onConnect(self, *args):
        """
        Callback function for handling MQTT connection.

        Args:
            *args: Variable length argument list.
        """
        if not self._connected:
            logging.info("Connected to MQTT broker [%s:%s] as %s", self._broker, self._port, self._client_id)
        self._connected = True

    def connect(self):
        """
        Connects to the MQTT broker.
        """
        self._client = mqtt.Client(self._client_id, clean_session=False)
        self._client.on_message = self._onMessage
        self._client.on_disconnect = self._onDisconnect
        self._client.on_connect = self._onConnect
        self._client.username_pw_set("backend", None)

        logging.debug("Connecting to MQTT broker [%s:%s] as %s...", self._broker, self._port, self._client_id)

        self._client.connect(self._broker, port=self._port)
        # Subscribe to all first-level subtopics of machine
        topic = self._topic + "/+"
        result = self._client.subscribe(topic, qos=1)
        if result[0] != mqtt.MQTT_ERR_SUCCESS:
            logging.error("Failure to subscribe topic [%s] : error %d", topic, result[0])
        else:
            logging.debug("Subscribed to topic [%s]", topic)

        self._client.loop_start()
        sleep(0.5)

    def setMessageCallback(self, callback: callable):
        """
        Sets the callback function for processing received messages.

        Args:
            callback (callable): The callback function.
        """
        self._messageCallback = callback

    def disconnect(self):
        """
        Disconnects from the MQTT broker.
        """
        self._client.unsubscribe(self._topic)
        self._client.loop_stop()
        self._client.disconnect()

    @property
    def connected(self):
        """
        Gets the connection status to the MQTT broker.

        Returns:
            bool: True if connected, False otherwise.
        """
        return self._connected

    def stats(self) -> dict:
        """
        Gets the statistics of the MQTT interface.

        Returns:
            dict: A dictionary containing the statistics.
        """
        import socket

        hostname = socket.gethostname()
        ipaddress = socket.gethostbyname(hostname)
        return {
            "Connected": self.connected,
            "Backend host": hostname,
            "Backend IP": ipaddress,
            "Received": self._msg_recv_count,
            "Sent": self._msg_send_count,
        }

    def publishStats(self):
        """
        Publishes the statistics to the MQTT broker.
        """
        self._publish(self._statsTopic, json.dumps(self.stats()))



"""Added replay columns for Use and Intervention

Revision ID: 3f3f601505a6
Revises: dfd2b0db0016
Create Date: 2024-05-30 23:09:40.723754

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "3f3f601505a6"
down_revision: Union[str, None] = "dfd2b0db0016"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("interventions", sa.Column("replay", sa.Boolean(), nullable=True))
    op.add_column("uses", sa.Column("replay", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("uses", "replay")
    op.drop_column("interventions", "replay")
    # ### end Alembic commands ###



import logging
import os
from logging.handlers import RotatingFileHandler

import colorlog


def configure_logger(loglevel=logging.DEBUG):
    """
    Configures the logger with the specified log level.

    Args:
        loglevel (int): The log level to set for the logger. Defaults to logging.DEBUG.

    Returns:
        None
    """
    # Create a logger object
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # Create a formatter for the logs
    formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")

    # Ensure the log directory is created
    log_dir = os.path.expanduser("~/log")
    os.makedirs(log_dir, exist_ok=True)

    # Create a rotating file handler with a maximum size of 1 MB
    log_file = os.path.join(log_dir, "log.txt")
    file_handler = RotatingFileHandler(log_file, maxBytes=1000000, backupCount=1, encoding="latin-1")
    file_handler.setFormatter(formatter)

    # Add the file handler to the logger object
    logger.addHandler(file_handler)

    # Create a stream handler to log to console
    console_handler = logging.StreamHandler()
    formatter2 = colorlog.ColoredFormatter(
        "%(log_color)s%(asctime)s %(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )
    console_handler.setFormatter(formatter2)

    # Add the stream handler to the logger object
    logger.addHandler(console_handler)



""" Test the MQTT interface and the message mapper """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import unittest

from FabOMatic.mqtt.mqtt_types import (
    MachineQuery,
    MachineResponse,
    SimpleResponse,
    StartUseQuery,
    UserQuery,
    UserResponse,
    EndUseQuery,
    RegisterMaintenanceQuery,
    AliveQuery,
)
from FabOMatic.mqtt.MQTTInterface import MQTTInterface
from FabOMatic.logic.MsgMapper import MsgMapper
from tests.common import TEST_SETTINGS_PATH, get_simple_db


class TestMQTT(unittest.TestCase):
    def test_json_deserialize(self):
        json_user_query = '{"action": "checkuser", "uid": "1234567890"}'
        user_query = UserQuery.deserialize(json_user_query)
        self.assertEqual(user_query.uid, "1234567890")
        self.assertEqual(user_query.action, "checkuser")
        self.assertEqual(user_query.__class__, UserQuery)

        json_machine_query = '{"action": "checkmachine"}'
        machine_query = MachineQuery.deserialize(json_machine_query)
        self.assertEqual(machine_query.action, "checkmachine")
        self.assertEqual(machine_query.__class__, MachineQuery)

        json_start_use_query = '{"action": "startuse", "uid": "1234", "replay":true}'
        start_use_query = StartUseQuery.deserialize(json_start_use_query)
        self.assertEqual(start_use_query.uid, "1234")
        self.assertEqual(start_use_query.action, "startuse")
        self.assertEqual(start_use_query.__class__, StartUseQuery)
        self.assertEqual(start_use_query.replay, True)

        json_stop_use_query = '{"action": "stopuse", "uid": "1234", "duration": 123, "replay":true}'
        stop_use_query = EndUseQuery.deserialize(json_stop_use_query)
        self.assertEqual(stop_use_query.uid, "1234")
        self.assertEqual(stop_use_query.duration, 123)
        self.assertEqual(stop_use_query.action, "stopuse")
        self.assertEqual(stop_use_query.replay, True)

        json_RegisterMaintenanceQuery = '{"action": "maintenance", "uid": "1234", "replay":true}'
        register_maintenance_query = RegisterMaintenanceQuery.deserialize(json_RegisterMaintenanceQuery)
        self.assertEqual(register_maintenance_query.uid, "1234")
        self.assertEqual(register_maintenance_query.action, "maintenance")
        self.assertEqual(register_maintenance_query.replay, True)
        self.assertEqual(register_maintenance_query.__class__, RegisterMaintenanceQuery)

        json_alive_query = '{"action": "alive", "ip": "1.2.3.4", "version": "1.2.3", "serial": "1234", "heap": 300000}'
        alive_query = AliveQuery.deserialize(json_alive_query)
        self.assertEqual(alive_query.action, "alive")
        self.assertEqual(alive_query.__class__, AliveQuery)

    def test_json_serialize(self):
        response = UserResponse(True, True, "user name", 2, False)
        json_response = response.serialize()
        self.assertEqual(
            json_response,
            '{"request_ok": true, "is_valid": true, "name": "user name", "missing_auth": false, "level": 2}',
        )

        response = MachineResponse(True, True, False, True, "Machine", 1, 120, 2, "Description")
        json_response = response.serialize()
        self.assertEqual(
            json_response,
            '{"request_ok": true, "is_valid": true, "maintenance": false, '
            + '"allowed": true, "name": "Machine", "logoff": 120, "type": 1, "grace": 2, "description": "Description"}',
        )

        response = SimpleResponse(True)
        json_response = response.serialize()
        self.assertEqual(json_response, '{"request_ok": true, "message": ""}')

    def test_init(self):
        d = MQTTInterface(TEST_SETTINGS_PATH)
        self.assertIsNotNone(d)
        d.connect()
        self.assertTrue(d.connected)

    def test_alive(self):
        CARD_UUID = "1234567890"
        NB_TESTS = 10

        db = get_simple_db()
        session = db.getSession()
        mqtt = MQTTInterface(TEST_SETTINGS_PATH)
        mapper = MsgMapper(mqtt, db)
        mapper.registerHandlers()
        mqtt.connect()
        self.assertTrue(mqtt.connected)
        user_repo = db.getUserRepository(session)

        # Create card if not presents
        user = user_repo.getUserByCardUUID(CARD_UUID)
        if user is None:
            user = user_repo.get_all()[0]
            user.card_UUID = CARD_UUID
            user_repo.update(user)

        # Generate all possible messages from the boards on all machines
        for _ in range(NB_TESTS):
            for mac in db.getMachineRepository(session).get_all():
                self.assertTrue(mqtt.publishQuery(mac.machine_id, '{"action": "alive"}'))
                self.assertTrue(mqtt.publishQuery(mac.machine_id, '{"action": "checkmachine"}'))
                self.assertTrue(
                    mqtt.publishQuery(mac.machine_id, '{"action": "checkuser", "uid": "' + CARD_UUID + '"}')
                )
                self.assertTrue(
                    mqtt.publishQuery(mac.machine_id, '{"action": "startuse", "uid": "' + CARD_UUID + '"}')
                )
                self.assertTrue(
                    mqtt.publishQuery(
                        mac.machine_id, '{"action": "stopuse", "uid": "' + CARD_UUID + '", "duration": 123}'
                    )
                )
                self.assertTrue(
                    mqtt.publishQuery(mac.machine_id, '{"action": "maintenance", "uid": "' + CARD_UUID + '"}')
                )


if __name__ == "__main__":
    unittest.main()



"""Added Boards table

Revision ID: 123b2cca325f
Revises: 88feb1b31a1f
Create Date: 2024-02-20 23:00:47.808587

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '123b2cca325f'
down_revision: Union[str, None] = '88feb1b31a1f'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('boards',
    sa.Column('board_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.Column('ip_address', sa.String(), nullable=False),
    sa.Column('fw_version', sa.String(), nullable=False),
    sa.Column('last_seen', sa.Float(), nullable=False),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.PrimaryKeyConstraint('board_id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('boards')
    # ### end Alembic commands ###



from flask import session, redirect, url_for
from .webapplication import app


@app.route("/language/<lang>")
def set_language(lang=None):
    session["language"] = lang
    return redirect(url_for("about"))



""" This module contains the routes for the interventions. """

# pylint: disable=C0116

from datetime import datetime
from time import time

from flask import render_template, request, redirect, url_for
from flask_login import login_required
from flask_babel import gettext
from FabOMatic.database.models import Intervention, Machine, Maintenance, User
from .webapplication import DBSession, app, excel


@app.route("/interventions")
@login_required
def view_interventions():
    session = DBSession()
    users = session.query(User).filter_by(deleted=False).order_by(User.user_id).all()
    machines = session.query(Machine).order_by(Machine.machine_id).all()

    search_user = request.args.get("search_user_id")
    search_machine = request.args.get("search_machine_id")

    query = session.query(Intervention)

    if search_user and search_user.isdigit():
        search_user = int(search_user)
        query = query.filter(Intervention.user_id == search_user)
    if search_machine and search_machine.isdigit():
        search_machine = int(search_machine)
        query = query.filter(Intervention.machine_id == search_machine)

    interventions = query.limit(500).all()
    return render_template(
        "view_interventions.html",
        users=users,
        machines=machines,
        interventions=interventions,
        search_user_id=search_user,
        search_machine_id=search_machine,
    )


@app.route("/interventions/add", methods=["GET", "POST"])
@login_required
def add_intervention():
    session = DBSession()
    machines = session.query(Machine).order_by(Machine.machine_id).all()
    users = session.query(User).order_by(User.user_id).all()
    maintenances = session.query(Maintenance).order_by(Maintenance.maintenance_id).all()

    if request.method == "POST":
        maintenance_id = request.form["maintenance_id"]
        machine_id = request.form["machine_id"]
        user_id = request.form["user_id"]
        timestamp = time()

        intervention = Intervention(
            maintenance_id=maintenance_id, machine_id=machine_id, user_id=user_id, timestamp=timestamp
        )

        session.add(intervention)
        session.commit()

        return redirect(url_for("view_interventions"))
    else:
        return render_template("add_intervention.html", machines=machines, users=users, maintenances=maintenances)


@app.route("/interventions/edit/<int:intervention_id>", methods=["GET", "POST"])
@login_required
def edit_intervention(intervention_id):
    session = DBSession()
    intervention = session.query(Intervention).get(intervention_id)
    machines = session.query(Machine).order_by(Machine.machine_id).all()
    users = session.query(User).filter_by(deleted=False).order_by(User.user_id).all()
    maintenances = session.query(Maintenance).order_by(Maintenance.maintenance_id).all()

    if request.method == "POST":
        intervention.maintenance_id = request.form["maintenance_id"]
        intervention.machine_id = request.form["machine_id"]
        intervention.user_id = request.form["user_id"]
        try:
            timestamp = datetime.strptime(request.form["timestamp"], "%Y-%m-%dT%H:%M")
            timestamp = timestamp.timestamp()
        except ValueError:
            timestamp = None

        intervention.timestamp = timestamp
        session.add(intervention)
        session.commit()

        return redirect(url_for("view_interventions"))
    else:
        return render_template(
            "edit_intervention.html",
            intervention=intervention,
            machines=machines,
            users=users,
            maintenances=maintenances,
        )


@app.route("/interventions/delete/<int:intervention_id>", methods=["GET", "POST"])
@login_required
def delete_intervention(intervention_id):
    session = DBSession()
    intervention = session.query(Intervention).get(intervention_id)

    if not intervention:
        return gettext("Intervention not found"), 404

    if request.method == "POST":
        session.delete(intervention)
        session.commit()

        return redirect(url_for("view_interventions"))

    return render_template("delete_intervention.html", intervention=intervention)


@app.route("/interventions/export", methods=["GET"])
def interventions_export():
    session = DBSession()
    return excel.make_response_from_tables(session, [Intervention, User, Machine], "xlsx", file_name="interventions")



""" This module contains the web application and common functions. """

# pylint: disable=C0116

from datetime import datetime
import os
from time import time

from flask import Flask, render_template, request, send_from_directory, g, session
from flask_login import login_required
from flask_babel import gettext
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from FabOMatic.database.models import Base
from FabOMatic.database.DatabaseBackend import getSetting, getDatabaseUrl
from FabOMatic.database.repositories import MachineRepository
from flask_babel import Babel
import flask_excel as excel

MODULE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
FLASK_TEMPLATES_FOLDER = os.path.join(MODULE_DIR, "flask_app", "templates")
FLASK_STATIC_FOLDER = os.path.join(MODULE_DIR, "flask_app", "static")
UPLOAD_FOLDER = os.path.join(MODULE_DIR, "flask_app", "upload")
ALLOWED_EXTENSIONS = {"txt", "pdf", "docx"}

app = Flask(__name__, template_folder=FLASK_TEMPLATES_FOLDER, static_folder=FLASK_STATIC_FOLDER)
app.config["SECRET_KEY"] = getSetting("web", "secret_key")

engine = create_engine(getDatabaseUrl(), echo=False)
Base.metadata.bind = engine
DBSession = sessionmaker(bind=engine)

# Init various extensions

excel.init_excel(app)


def get_locale():
    # if the user has set up the language manually it will be stored in the session,
    # so we use the locale from the user settings
    try:
        language = session["language"]
    except KeyError:
        language = None
    if language is not None:
        return language

    translations = [str(translation) for translation in babel.list_translations()]
    return request.accept_languages.best_match(translations)


def get_timezone():
    user = getattr(g, "user", None)
    if user is not None:
        return user.timezone


babel = Babel(
    app,
    locale_selector=get_locale,
    timezone_selector=get_timezone,
    default_timezone="Europe/Rome",
    default_locale="it",
    default_translation_directories=os.path.join(MODULE_DIR, "translations"),
)


def allowed_file(filename):
    return "." in filename and filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS


def timestamp_to_datetime(value):
    return datetime.fromtimestamp(value)


# Add the filter to Jinja2's environment
app.jinja_env.filters["timestamp_to_datetime"] = timestamp_to_datetime


# Custom filter for datetime formatting
@app.template_filter("datetimeformat")
def datetimeformat(value, format="%b %d, %Y %I:%M %p"):
    if value is None:
        return "-"
    if not isinstance(value, datetime):
        value = datetime.fromtimestamp(value)
    return value.strftime(format)


@app.template_filter("format_hours")
def format_hours(value):
    if value is None:
        return "-"

    hours = int(value)
    minutes = int((value % 1) * 60)

    if hours > 1:
        hour_text = gettext("hours")
    else:
        hour_text = gettext("hour")

    if minutes > 1:
        minutes_text = gettext("minutes")
    else:
        minutes_text = gettext("minute")

    if hours == 0:
        if minutes == 0:
            return "-"
        return f"{minutes} {minutes_text}"

    return f"{hours} {hour_text} {minutes} {minutes_text}"


@app.template_filter("time_since")
def time_since(dt):
    now = datetime.now()

    if not isinstance(dt, datetime):
        try:
            value = int(dt)
            dt = datetime.fromtimestamp(value)
        except ValueError:
            dt = datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

    diff = now - dt
    seconds = int(abs(diff.total_seconds()))

    if seconds < 60:
        if seconds > 1:
            return f"{seconds} " + gettext("seconds ago")
        else:
            return f"{seconds} " + gettext("second ago")
    elif seconds < 3600:
        if seconds // 60 > 1:
            return f"{seconds // 60} " + gettext("minutes ago")
        else:
            return f"{seconds // 60} " + gettext("minute ago")
    elif seconds < 86400:
        if seconds // 3600 > 1:
            return f"{seconds // 3600} " + gettext("hours ago")
        else:
            return f"{seconds // 3600} " + gettext("hour ago")
    else:
        if seconds // 86400 > 1:
            return f"{seconds // 86400} " + gettext("days ago")

    return f"{seconds // 86400} " + gettext("day ago")


# Define routes
@app.route("/")
def index():
    return render_template("index.html")


@app.route("/about")
def about():
    with DBSession() as session:
        machine_repo = MachineRepository(session)
        machines = machine_repo.get_all()
        for mac in machines:
            setattr(mac, "maintenance_needed", machine_repo.getMachineMaintenanceNeeded(mac.machine_id)[0])
            if mac.last_seen is None:
                setattr(mac, "online", False)
            else:
                setattr(mac, "online", time() - mac.last_seen < 180)
        return render_template("about.html", machines=machines)



from . import (
    webapplication,
    authentication,
    routes_authorizations,
    routes_interventions,
    routes_languages,
    routes_machines,
    routes_machinetypes,
    routes_maintenance,
    routes_roles,
    routes_users,
    routes_uses,
    routes_system,
)



""" Test the backend module. """

# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring

import unittest

from FabOMatic.__main__ import Backend
from tests.common import TEST_SETTINGS_PATH


class TestBackend(unittest.TestCase):
    def test_backend(self):
        backend = Backend(TEST_SETTINGS_PATH)
        self.assertTrue(backend.connect(), "Failed to connect the first time")
        backend.disconnect()
        self.assertTrue(backend.connect(), "Failed to connect the second time")
        backend.publishStats()



"""Initial database creation

Revision ID: 2dc275252e0e
Revises: 
Create Date: 2024-01-28 16:25:31.644671

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '2dc275252e0e'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('machine_types',
    sa.Column('type_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('type_name', sa.String(), nullable=False),
    sa.Column('type_timeout_min', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('type_id'),
    sa.UniqueConstraint('type_name')
    )
    op.create_index('idx_machine_types_type_name_unique', 'machine_types', ['type_name'], unique=True)
    op.create_table('roles',
    sa.Column('role_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('role_name', sa.String(), nullable=False),
    sa.Column('authorize_all', sa.Boolean(), nullable=False),
    sa.Column('reserved', sa.Boolean(), nullable=False),
    sa.Column('maintenance', sa.Boolean(), nullable=False),
    sa.Column('backend_admin', sa.Boolean(), nullable=False),
    sa.PrimaryKeyConstraint('role_id'),
    sa.UniqueConstraint('role_name')
    )
    op.create_index('idx_roles_role_name_unique', 'roles', ['role_name'], unique=True)
    op.create_table('machines',
    sa.Column('machine_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('machine_name', sa.String(), nullable=False),
    sa.Column('machine_type_id', sa.Integer(), nullable=True),
    sa.Column('machine_hours', sa.Float(), nullable=False),
    sa.Column('blocked', sa.Boolean(), nullable=False),
    sa.Column('last_seen', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['machine_type_id'], ['machine_types.type_id'], ),
    sa.PrimaryKeyConstraint('machine_id'),
    sa.UniqueConstraint('machine_name')
    )
    op.create_index('idx_machines_machine_name_unique', 'machines', ['machine_name'], unique=True)
    op.create_table('users',
    sa.Column('user_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('surname', sa.String(), nullable=False),
    sa.Column('role_id', sa.Integer(), nullable=False),
    sa.Column('card_UUID', sa.String(), nullable=True),
    sa.Column('disabled', sa.Boolean(), nullable=False),
    sa.Column('deleted', sa.Boolean(), nullable=False),
    sa.Column('password_hash', sa.String(length=128), nullable=True),
    sa.Column('email', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['role_id'], ['roles.role_id'], ),
    sa.PrimaryKeyConstraint('user_id'),
    sa.UniqueConstraint('card_UUID')
    )
    op.create_index('idx_users_card_UUID_unique', 'users', ['card_UUID'], unique=True)
    op.create_table('authorizations',
    sa.Column('authorization_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.user_id'], ),
    sa.PrimaryKeyConstraint('authorization_id'),
    sa.UniqueConstraint('user_id', 'machine_id', name='uix_1')
    )
    op.create_table('maintenances',
    sa.Column('maintenance_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('hours_between', sa.Float(), nullable=False),
    sa.Column('description', sa.String(), nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.Column('attachment', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.PrimaryKeyConstraint('maintenance_id')
    )
    op.create_table('unknown_cards',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('card_UUID', sa.String(), nullable=False),
    sa.Column('timestamp', sa.Float(), nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('uses',
    sa.Column('use_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.Column('start_timestamp', sa.Float(), nullable=False),
    sa.Column('last_seen', sa.Float(), nullable=False),
    sa.Column('end_timestamp', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.user_id'], ),
    sa.PrimaryKeyConstraint('use_id')
    )
    op.create_table('interventions',
    sa.Column('intervention_id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('maintenance_id', sa.Integer(), nullable=False),
    sa.Column('machine_id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('timestamp', sa.Float(), nullable=False),
    sa.ForeignKeyConstraint(['machine_id'], ['machines.machine_id'], ),
    sa.ForeignKeyConstraint(['maintenance_id'], ['maintenances.maintenance_id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.user_id'], ),
    sa.PrimaryKeyConstraint('intervention_id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('interventions')
    op.drop_table('uses')
    op.drop_table('unknown_cards')
    op.drop_table('maintenances')
    op.drop_table('authorizations')
    op.drop_index('idx_users_card_UUID_unique', table_name='users')
    op.drop_table('users')
    op.drop_index('idx_machines_machine_name_unique', table_name='machines')
    op.drop_table('machines')
    op.drop_index('idx_roles_role_name_unique', table_name='roles')
    op.drop_table('roles')
    op.drop_index('idx_machine_types_type_name_unique', table_name='machine_types')
    op.drop_table('machine_types')
    # ### end Alembic commands ###



""" This module provides the MsgMapper class"""

import logging

from FabOMatic.database.DatabaseBackend import DatabaseBackend
from FabOMatic.mqtt import MQTTInterface
from FabOMatic.mqtt.mqtt_types import (
    UserQuery,
    StartUseQuery,
    InUseQuery,
    EndUseQuery,
    RegisterMaintenanceQuery,
    AliveQuery,
    MachineQuery,
    BaseJson,
    SimpleResponse,
)

from .MachineLogic import MachineLogic


class MsgMapper:
    """This class provides the handlers that incoming parsed MQTT message
    to the machine_logic instance, and returns the response as a string ."""

    def __init__(self, mqtt: MQTTInterface, db: DatabaseBackend):
        MachineLogic.database = db
        self._mqtt = mqtt
        self._db = db
        self._machines = {}
        self._handlers = {}

    def getMachineLogic(self, mid: int) -> MachineLogic | None:
        """
        Retrieves the MachineLogic object associated with the given machine ID.

        Args:
            machineId (str): The ID of the machine.

        Returns:
            MachineLogic | None: The MachineLogic object if found, None otherwise.
        """
        if mid not in self._machines:
            try:
                self._machines[mid] = MachineLogic(mid)
                logging.info(
                    "Created MachineLogic instance for machine %s, %d machines total.", mid, len(self._machines)
                )
            except Exception as e:
                logging.error("MachineLogic creation exception %s", str(e))
                return None
        return self._machines[mid]

    def handleUserQuery(self, machine_logic: MachineLogic, userquery: UserQuery) -> str:
        response = machine_logic.isAuthorized(userquery.uid)
        logging.debug("User query: %s -> response: %s", userquery.toJSON(), response.serialize())
        return response.serialize()

    def handleStartUseQuery(self, machine_logic: MachineLogic, startUse: StartUseQuery) -> str:
        response = machine_logic.startUse(startUse.uid, startUse.replay)
        logging.info(
            "[Machine %d] Start use query: %s -> response: %s",
            machine_logic.getMachineId(),
            startUse.toJSON(),
            response.serialize(),
        )
        return response.serialize()

    def handleInUseQuery(self, machine_logic: MachineLogic, inUse: InUseQuery) -> str:
        response = machine_logic.inUse(inUse.uid, inUse.duration)
        logging.info(
            "[Machine %d] In use query: %s -> response: %s",
            machine_logic.getMachineId(),
            inUse.toJSON(),
            response.serialize(),
        )
        return response.serialize()

    def handleEndUseQuery(self, machine_logic: MachineLogic, stopUse: EndUseQuery) -> str:
        response = machine_logic.endUse(stopUse.uid, stopUse.duration, stopUse.replay)
        logging.info(
            "[Machine %d] End use query: %s -> response: %s",
            machine_logic.getMachineId(),
            stopUse.toJSON(),
            response.serialize(),
        )
        return response.serialize()

    def handleMaintenanceQuery(self, machine_logic: MachineLogic, maintenance: RegisterMaintenanceQuery) -> str:
        response = machine_logic.registerMaintenance(maintenance.uid, maintenance.replay)
        logging.info(
            "[Machine %d] Start use query: %s -> response: %s",
            machine_logic.getMachineId(),
            maintenance.toJSON(),
            response.serialize(),
        )
        return response.serialize()

    def handleAliveQuery(self, machine_logic: MachineLogic, alive: AliveQuery) -> str:
        """
        Handles an alive query message.

        Args:
            machine_logic (MachineLogic): The machine logic instance.
            alive (AliveQuery): The alive query message.

        Returns:
            str: None
        """
        machine_logic.machineAlive(alive)
        logging.debug("Alive query: %s", alive.toJSON())
        return None

    def handleMachineQuery(self, machine_logic: MachineLogic, machineQuery: MachineQuery) -> str:
        """
        Handles a machine query and returns the serialized machine status.

        Args:
            machine_logic (MachineLogic): The machine logic object.
            machineQuery (MachineQuery): The machine query object.

        Returns:
            str: The serialized machine status.
        """
        status = machine_logic.machineStatus()
        logging.debug("Machine query: %s -> response: %s", machineQuery.toJSON(), status.serialize())
        return status.serialize()

    def messageReceived(self, machine: int, query: BaseJson) -> bool:
        """This function is called when a message is received from the MQTT broker.
        It calls the appropriate handler for the message type."""

        if type(query) not in self._handlers:
            logging.warning(f"No handler for query {query} on machine {machine}")
            return False

        machine_logic = self.getMachineLogic(machine)
        if machine_logic is None:
            logging.error(f"Failed to create MachineLogic instance for machine {machine}")
            response = SimpleResponse(False, "Invalid machine ID").serialize()
            if not self._mqtt.publishReply(machine, response):
                logging.error(f"Failed to publish response for machine {machine} to MQTT broker: {response}")
            return False

        response = self._handlers[type(query)](machine_logic, query)

        if response is not None:
            if not self._mqtt.publishReply(machine, response):
                logging.error(f"Failed to publish response for machine {machine} to MQTT broker: {response}")
                return False
        else:
            logging.warning(f"Machine {machine} query: {query.toJSON()} -> no response")
            return False

        return True

    def registerHandlers(self):
        """This function registers the handlers for the different message types from the boards."""
        self._setHandler(AliveQuery, self.handleAliveQuery)
        self._setHandler(MachineQuery, self.handleMachineQuery)
        self._setHandler(UserQuery, self.handleUserQuery)
        self._setHandler(StartUseQuery, self.handleStartUseQuery)
        self._setHandler(InUseQuery, self.handleInUseQuery)
        self._setHandler(EndUseQuery, self.handleEndUseQuery)
        self._setHandler(RegisterMaintenanceQuery, self.handleMaintenanceQuery)
        self._mqtt.setMessageCallback(self.messageReceived)

    def _setHandler(self, query: type, handler: callable):
        self._handlers[query] = handler
